@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology.
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,8 +32,6 @@
 
 using namespace std;
 
-typedef complex<double> cdouble;
-
 namespace meep {
 
 #ifdef HAVE_MPB
@@ -158,7 +156,7 @@ static int pmod(int n, int modulus) {
 /*******************************************************************/
 complex<double> eigenmode_amplitude(void *vedata, const vec &p, component c) {
   eigenmode_data *edata = (eigenmode_data *)vedata;
-  if (!edata || !(edata->mdata)) abort("%s:%i: internal error", __FILE__, __LINE__);
+  if (!edata || !(edata->mdata)) meep::abort("%s:%i: internal error", __FILE__, __LINE__);
 
   int *n = edata->n;
   double *s = edata->s;
@@ -193,7 +191,7 @@ complex<double> eigenmode_amplitude(void *vedata, const vec &p, component c) {
       cdata = (complex<mpb_real> *)edata->fft_data_H;
       data = cdata + 2;
       break;
-    default: abort("invalid component in eigenmode_amplitude");
+    default: meep::abort("invalid component in eigenmode_amplitude");
   };
 
   int nx = n[0];
@@ -294,6 +292,18 @@ static int nextpow2357(int n) {
   }
 }
 
+void special_kz_phasefix(eigenmode_data *edata, bool phase_flip) {
+  size_t n = edata->n[0] * edata->n[1] * edata->n[2];
+  complex<mpb_real> *E = (complex<mpb_real> *)edata->fft_data_E;
+  complex<mpb_real> *H = (complex<mpb_real> *)edata->fft_data_H;
+  complex<mpb_real> im(0, phase_flip ? -1 : 1);
+  for (size_t i = 0; i < n; ++i) {
+    E[3*i + 2] *= im; // Ez
+    H[3*i + 0] *= im; // Hx
+    H[3*i + 1] *= im; // Hy
+  }
+}
+
 /****************************************************************/
 /* call MPB to get the band_numth eigenmode at freq frequency.  */
 /*                                                              */
@@ -326,7 +336,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   // k-vector to be the (real part of the) bloch vector in that direction.
   vec kpoint(_kpoint);
   LOOP_OVER_DIRECTIONS(v.dim, dd) {
-    if (float(eig_vol.in_direction(dd)) == float(v.in_direction(dd)))
+    if (dd != d && float(eig_vol.in_direction(dd)) == float(v.in_direction(dd)))
       if (boundaries[High][dd] == Periodic && boundaries[Low][dd] == Periodic)
         kpoint.set_direction(dd, real(k[dd]));
   }
@@ -335,8 +345,8 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
 
   // special case: 2d cell in x and y with non-zero kz
   if ((eig_vol.dim == D3) && (float(v.in_direction(Z)) == float(1 / a)) &&
-      (boundaries[High][Z] == Periodic && boundaries[Low][Z] == Periodic) && (kpoint.z() == 0) &&
-      (real(k[Z]) != 0)) {
+      (boundaries[High][Z] == Periodic && boundaries[Low][Z] == Periodic) &&
+      (kpoint.z() == 0) && (real(k[Z]) != 0)) {
     kpoint.set_direction(Z, real(k[Z]));
     empty_dim[2] = true;
   }
@@ -351,12 +361,12 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   double match_tol = eigensolver_tol * 10;
 
   if ((d == NO_DIRECTION && abs(_kpoint) == 0) || coordinate_mismatch(gv.dim, d))
-    abort("invalid direction in add_eigenmode_source");
+    meep::abort("invalid direction in add_eigenmode_source");
   if (where.dim != gv.dim || eig_vol.dim != gv.dim)
-    abort("invalid volume dimensionality in add_eigenmode_source");
+    meep::abort("invalid volume dimensionality in add_eigenmode_source");
 
   if (!eig_vol.contains(where))
-    abort("invalid grid_volume in add_eigenmode_source (WHERE must be in EIG_VOL)");
+    meep::abort("invalid grid_volume in add_eigenmode_source (WHERE must be in EIG_VOL)");
 
   switch (gv.dim) {
     case D3:
@@ -386,7 +396,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
       kcart[2] = kpoint.in_direction(Z);
       empty_dim[0] = empty_dim[1] = true;
       break;
-    default: abort("unsupported dimensionality in add_eigenmode_source");
+    default: meep::abort("unsupported dimensionality in add_eigenmode_source");
   }
 
   double kcart_len = sqrt(dot_product(kcart, kcart));
@@ -422,7 +432,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   maxwell_data *mdata;
   if (!user_mdata || *user_mdata == NULL) {
     mdata = create_maxwell_data(n[0], n[1], n[2], &local_N, &N_start, &alloc_N, band_num, band_num);
-    if (local_N != n[0] * n[1] * n[2]) abort("MPI version of MPB library not supported");
+    if (local_N != n[0] * n[1] * n[2]) meep::abort("MPI version of MPB library not supported");
 
     meep_mpb_eps_data eps_data;
     eps_data.s = s;
@@ -440,7 +450,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
     // then, synchronize the data
     eps_data.ncache = eps_data.icache; // actual amount of cached data
     double *summed_cache = (double *)malloc(sizeof(double) * 6 * eps_data.ncache);
-    am_now_working_on(MpiTime);
+    am_now_working_on(MpiAllTime);
     sum_to_all(eps_data.cache, summed_cache, eps_data.ncache * 6);
     finished_working();
     free(eps_data.cache);
@@ -464,7 +474,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
     alloc_N = mdata->alloc_N;
   }
 
-  if (check_maxwell_dielectric(mdata, 0)) abort("invalid dielectric function for MPB");
+  if (check_maxwell_dielectric(mdata, 0)) meep::abort("invalid dielectric function for MPB");
 
   double kmatch;
   if (d == NO_DIRECTION) {
@@ -509,18 +519,19 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
     H.p = band_num;
     H.c = 2;
     band_num -= maxwell_zero_k_num_const_bands(H, mdata);
-    if (band_num == 0) abort("zero-frequency bands at k=0 are ill-defined");
+    if (band_num == 0) meep::abort("zero-frequency bands at k=0 are ill-defined");
   }
 
   evectmatrix H = create_evectmatrix(n[0] * n[1] * n[2], 2, band_num, local_N, N_start, alloc_N);
   /* initialize H to pseudorandom values on the master process; on other
      processes we get the value via broadcast() below */
-  if (am_master())
+  if (am_master()) {
     set_random_seed(314159);
     for (int i = 0; i < H.n * H.p; ++i) {
       ASSIGN_SCALAR(H.data[i], uniform_random(-1,1), uniform_random(-1,1));
     }
     restore_random_seed();
+  }
 
   mpb_real *eigvals = new mpb_real[band_num];
   int num_iters;
@@ -605,16 +616,29 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
     int m = 0;
     LOOP_OVER_DIRECTIONS(v.dim, dd) {
       m = dp->get_g()[dd - X];
-      ktmp = kpoint.in_direction(dd);
-      if (eig_vol.in_direction(dd) != 0)
-        ktmp += m/eig_vol.in_direction(dd);
-      k2sum += ktmp*ktmp;
+      if (eig_vol.in_direction(dd) != 0) {
+        ktmp = kpoint.in_direction(dd) + m/eig_vol.in_direction(dd);
+        k2sum += ktmp*ktmp;
+      }
+    }
+    if (((v.dim == D3) && (float(v.in_direction(Z)) == float(1 / a)) &&
+        (boundaries[High][Z] == Periodic && boundaries[Low][Z] == Periodic) && (real(k[Z]) != 0)) ||
+        ((v.dim == D2) && (real(k[Z]) != 0))) {
+      k2sum += k[Z]*k[Z];
     }
 
     // compute kperp (if it is non evanescent) OR
     // frequency from kperp^2 and sum of (kparallel+G)^2
     {
       direction dd = eig_vol.normal_direction();
+      if (eig_vol.dim == D3 && empty_dim[2]) {
+        volume eig_vol_2d(D2);
+        eig_vol_2d.set_direction_min(X, eig_vol.in_direction_min(X));
+        eig_vol_2d.set_direction_min(Y, eig_vol.in_direction_min(Y));
+        eig_vol_2d.set_direction_max(X, eig_vol.in_direction_max(X));
+        eig_vol_2d.set_direction_max(Y, eig_vol.in_direction_max(Y));
+        dd = eig_vol_2d.normal_direction();
+      }
       if (match_frequency) {
         vec cen = eig_vol.center();
         double nn = sqrt(real(get_eps(cen, frequency)) * real(get_mu(cen, frequency)));
@@ -654,7 +678,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   for (int i = 0; i < 3; ++i)
     destroy_evectmatrix(W[i]);
 
-  am_now_working_on(MpiTime);
+  am_now_working_on(MpiAllTime);
   /* We only run MPB eigensolver on the master process to avoid
      any possibility of inconsistent mode solutions (#568) */
   eigval = broadcast(0, eigval);
@@ -721,8 +745,8 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   // d_from_H actually computes -frequency*D (see mpb/src/maxwell/maxwell_op.c),
   // so we need to divide the E-field amplitudes by -frequency; we also take this
   // opportunity to rescale the overall E and H amplitudes to yield unit power flux.
-  double scale = -1.0 / frequency, factor = 2.0 / sqrt(vgrp);
-  cdouble *efield = (cdouble *)fft_data_E, *hfield = (cdouble *)(mdata->fft_data);
+  double scale = -1.0 / frequency, factor = 2.0 / sqrt(fabs(vgrp));
+  complex<double> *efield = (complex<double> *)fft_data_E, *hfield = (complex<double> *)(mdata->fft_data);
   for (int n = 0; n < NFFT; ++n) {
     efield[n] *= factor * scale;
     hfield[n] *= factor;
@@ -810,13 +834,17 @@ void fields::add_eigenmode_source(component c0, const src_time &src, direction d
                                       NULL, NULL, dp);
   finished_working();
 
+  if (is_real && beta != 0) special_kz_phasefix(global_eigenmode_data, true /* phase_flip */);
+
+  if (global_eigenmode_data == NULL) meep::abort("MPB could not find the eigenmode");
+
   /* add_volume_source amp_fun coordinates are relative to where.center();
      this is not the default in get_eigenmode because where-relative coordinates
      are not used elsewhere, e.g. in getting mode coefficients in dft.cpp. */
   global_eigenmode_data->center -= where.center();
 
   if (!global_eigenmode_data)
-    abort("eigenmode solver failed to find the requested mode; you may need to supply a better "
+    meep::abort("eigenmode solver failed to find the requested mode; you may need to supply a better "
           "guess for k");
 
   global_eigenmode_data->amp_func = A ? A : default_amp_func;
@@ -839,7 +867,7 @@ void fields::add_eigenmode_source(component c0, const src_time &src, direction d
             ? 0
             : where.in_direction(Y) == 0 ? 1 : where.in_direction(Z) == 0 ? 2 : -1;
     if (n == -1)
-      abort(
+      meep::abort(
           "can't determine source direction for non-empty source volume with NO_DIRECTION source");
   }
   int np1 = (n + 1) % 3;
@@ -892,7 +920,7 @@ void fields::get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, in
   int num_freqs = flux.freq.size();
   bool match_frequency = true;
   if (flux.use_symmetry && S.multiplicity() > 1 && parity == 0)
-    abort("flux regions for eigenmode projection with symmetry should be created by "
+    meep::abort("flux regions for eigenmode projection with symmetry should be created by "
           "add_mode_monitor()");
 
   vec kpoint(0.0, 0.0, 0.0); // default guess
@@ -922,6 +950,8 @@ void fields::get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, in
         continue;
       }
 
+      if (is_real && beta != 0) special_kz_phasefix((eigenmode_data *) mode_data, false /* phase_flip */);
+
       double vg = get_group_velocity(mode_data);
       vec kfound = get_k(mode_data);
       if (vgrp) vgrp[nb * num_freqs + nf] = vg;
@@ -938,15 +968,15 @@ void fields::get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, in
       /*--------------------------------------------------------------*/
       /*--------------------------------------------------------------*/
       /*--------------------------------------------------------------*/
-      cdouble mode_flux[2], mode_mode[2];
+      complex<double> mode_flux[2], mode_mode[2];
       get_mode_flux_overlap(mode_data, flux, nf, mode_flux);
       get_mode_mode_overlap(mode_data, mode_data, flux, mode_mode);
-      cdouble cplus = 0.5 * (mode_flux[0] + mode_flux[1]);
-      cdouble cminus = 0.5 * (mode_flux[0] - mode_flux[1]);
+      complex<double> cplus = 0.5 * (mode_flux[0] + mode_flux[1]);
+      complex<double> cminus = 0.5 * (mode_flux[0] - mode_flux[1]);
       /* MPB modes are normalized to unit power above, but we need to re-normalize here to have
          unit power as integrated on Meep's Yee grid and not on MPB's grid.  Thus, normfac differs
          from a constant factor only because of discretization effects. */
-      cdouble normfac = 0.5 * (mode_mode[0] + mode_mode[1]);
+      complex<double> normfac = 0.5 * (mode_mode[0] + mode_mode[1]);
       if (normfac == 0.0) normfac = 1.0;
       double csc = sqrt((flux.use_symmetry ? S.multiplicity() : 1.0) / abs(normfac));
       if (cscale)
@@ -982,7 +1012,7 @@ void *fields::get_eigenmode(double frequency, direction d, const volume where, c
   (void)kdom;
   (void)user_mdata;
   (void)dp;
-  abort("Meep must be configured/compiled with MPB for get_eigenmode");
+  meep::abort("Meep must be configured/compiled with MPB for get_eigenmode");
 }
 
 void fields::add_eigenmode_source(component c0, const src_time &src, direction d,
@@ -1004,7 +1034,7 @@ void fields::add_eigenmode_source(component c0, const src_time &src, direction d
   (void)amp;
   (void)A;
   (void)dp;
-  abort("Meep must be configured/compiled with MPB for add_eigenmode_source");
+  meep::abort("Meep must be configured/compiled with MPB for add_eigenmode_source");
 }
 
 void fields::get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, int *bands,
@@ -1029,7 +1059,7 @@ void fields::get_eigenmode_coefficients(dft_flux flux, const volume &eig_vol, in
   (void)cscale;
   (void)d;
   (void)dp;
-  abort("Meep must be configured/compiled with MPB for get_eigenmode_coefficients");
+  meep::abort("Meep must be configured/compiled with MPB for get_eigenmode_coefficients");
 }
 
 void destroy_eigenmode_data(void *vedata, bool destroy_mdata) {
