@@ -1,4 +1,4 @@
-/* Copyright (C) 2005-2021 Massachusetts Institute of Technology
+/* Copyright (C) 2005-2022 Massachusetts Institute of Technology
 %
 %  This program is free software; you can redistribute it and/or modify
 %  it under the terms of the GNU General Public License as published by
@@ -15,6 +15,8 @@
 %  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
 
+#include <cassert>
+#include <utility>
 #include <stdio.h>
 #include <stdlib.h>
 #include <math.h>
@@ -31,7 +33,7 @@ namespace meep {
 
 // this function is necessary to make equality commutative ... ugh
 bool src_times_equal(const src_time &t1, const src_time &t2) {
-  return t1.is_equal(t2) && t2.is_equal(t1);
+  return t1.is_equal(t2) && t2.is_equal(t1) && t1.is_integrated == t2.is_integrated;
 }
 
 src_time *src_time::add_to(src_time *others, src_time **added) const {
@@ -51,18 +53,27 @@ src_time *src_time::add_to(src_time *others, src_time **added) const {
 }
 
 double src_time::last_time_max(double after) {
-  after = max(last_time(), after);
+  after = std::max(last_time(), after);
   if (next)
     return next->last_time_max(after);
   else
     return after;
 }
 
-gaussian_src_time::gaussian_src_time(double f, double fwidth, double s) {
+// bandwidth (in frequency units, not angular frequency) of the
+// continuous Fourier transform of the Gaussian source function
+// when it has decayed by a tolerance tol below its peak value
+static double gaussian_bandwidth(double width) {
+  double tol = 1e-7;
+  return sqrt(-2.0 * log(tol)) / (width * pi);
+}
+
+gaussian_src_time::gaussian_src_time(double f, double my_fwidth, double s) {
   freq = f;
-  width = 1.0 / fwidth;
+  width = 1.0 / my_fwidth;
   peak_time = width * s;
   cutoff = width * s;
+  fwidth = gaussian_bandwidth(width);
 
   // this is to make last_source_time as small as possible
   while (exp(-cutoff * cutoff / (2 * width * width)) < 1e-100)
@@ -75,6 +86,7 @@ gaussian_src_time::gaussian_src_time(double f, double w, double st, double et) {
   width = w;
   peak_time = 0.5 * (st + et);
   cutoff = (et - st) * 0.5;
+  fwidth = gaussian_bandwidth(width);
 
   // this is to make last_source_time as small as possible
   while (exp(-cutoff * cutoff / (2 * width * width)) < 1e-100)
@@ -145,59 +157,32 @@ bool custom_src_time::is_equal(const src_time &t) const {
   const custom_src_time *tp = dynamic_cast<const custom_src_time *>(&t);
   if (tp)
     return (tp->start_time == start_time && tp->end_time == end_time && tp->func == func &&
-            tp->data == data);
+            tp->data == data && tp->freq == freq && tp->fwidth == fwidth);
   else
     return 0;
 }
 
 /*********************************************************************/
 
-src_vol::src_vol(component cc, src_time *st, size_t n, ptrdiff_t *ind, complex<double> *amps) {
-  c = cc;
-  if (is_D(c)) c = direction_component(Ex, component_direction(c));
-  if (is_B(c)) c = direction_component(Hx, component_direction(c));
-  t = st;
-  next = NULL;
-  npts = n;
-  index = ind;
-  A = amps;
+src_vol::src_vol(component cc, src_time *st, std::vector<ptrdiff_t> &&ind,
+                 std::vector<std::complex<double> > &&amps)
+    : c([](component c) -> component {
+        if (is_D(c)) c = direction_component(Ex, component_direction(c));
+        if (is_B(c)) c = direction_component(Hx, component_direction(c));
+        return c;
+      }(cc)),
+      src_t(st), index(std::move(ind)), amp(std::move(amps)) {
+  assert(index.size() == amp.size());
 }
 
-src_vol::src_vol(const src_vol &sv) {
-  c = sv.c;
-  t = sv.t;
-  npts = sv.npts;
-  index = new ptrdiff_t[npts];
-  A = new complex<double>[npts];
-  for (size_t j = 0; j < npts; j++) {
-    index[j] = sv.index[j];
-    A[j] = sv.A[j];
-  }
-  if (sv.next)
-    next = new src_vol(*sv.next);
-  else
-    next = NULL;
+bool src_vol::combinable(const src_vol &a, const src_vol &b) {
+  return (a.c == b.c) && (a.src_t == b.src_t) && (a.index == b.index);
 }
 
-src_vol *src_vol::add_to(src_vol *others) {
-  if (others) {
-    if (*this == *others) {
-      if (npts != others->npts)
-        abort("Cannot add grid_volume sources with different number of points\n");
-      /* Compare all of the indices...if this ever becomes too slow,
-         we can just compare the first and last indices. */
-      for (size_t j = 0; j < npts; j++) {
-        if (others->index[j] != index[j]) abort("Different indices\n");
-        others->A[j] += A[j];
-      }
-    }
-    else
-      others->next = add_to(others->next);
-    return others;
-  }
-  else {
-    next = others;
-    return this;
+void src_vol::add_amplitudes_from(const src_vol &other) {
+  assert(amp.size() == other.num_points());
+  for (size_t i = 0; i < amp.size(); ++i) {
+    amp[i] += other.amp[i];
   }
 }
 
@@ -267,8 +252,8 @@ static void src_vol_chunkloop(fields_chunk *fc, int ichunk, component c, ivec is
 
   size_t npts = 1;
   LOOP_OVER_DIRECTIONS(is.dim, d) { npts *= (ie.in_direction(d) - is.in_direction(d)) / 2 + 1; }
-  ptrdiff_t *index_array = new ptrdiff_t[npts];
-  complex<double> *amps_array = new complex<double>[npts];
+  std::vector<ptrdiff_t> index_array(npts);
+  std::vector<complex<double>> amps_array(npts);
 
   complex<double> amp = data->amp * conj(shift_phase);
 
@@ -297,33 +282,32 @@ static void src_vol_chunkloop(fields_chunk *fc, int ichunk, component c, ivec is
     index_array[idx_vol++] = idx;
   }
 
-  if (idx_vol > npts) abort("add_volume_source: computed wrong npts (%zd vs. %zd)", npts, idx_vol);
+  if (idx_vol > npts) meep::abort("add_volume_source: computed wrong npts (%zd vs. %zd)", npts, idx_vol);
 
-  src_vol *tmp = new src_vol(c, data->src, idx_vol, index_array, amps_array);
   field_type ft = is_magnetic(c) ? B_stuff : D_stuff;
-  fc->sources[ft] = tmp->add_to(fc->sources[ft]);
+  fc->add_source(ft, src_vol(c, data->src, std::move(index_array), std::move(amps_array)));
 }
 
 void fields::add_srcdata(struct sourcedata cur_data, src_time *src, size_t n, std::complex<double>* amp_arr){
   sources = src->add_to(sources, &src);
-  ptrdiff_t* index_arr = new ptrdiff_t[n];
-  std::complex<double>* amp_arr_copy = new std::complex<double>[n];
-  for (size_t i = 0; i < n; i++){
-    index_arr[i] = cur_data.idx_arr[i];
-    amp_arr_copy[i] = amp_arr[i];
-  }
+  std::vector<ptrdiff_t> index_arr(cur_data.idx_arr);
+  std::vector<std::complex<double>> amplitudes(amp_arr, amp_arr+n);
   component c = cur_data.near_fd_comp;
-  src_vol *tmp = new src_vol(c, src, n, index_arr, amp_arr_copy);
+
   field_type ft = is_magnetic(c) ? B_stuff : D_stuff;
-  if (0 > cur_data.fc_idx or cur_data.fc_idx >= num_chunks) abort("fields chunk index out of range");
+  if (0 > cur_data.fc_idx or cur_data.fc_idx >= num_chunks) meep::abort("fields chunk index out of range");
   fields_chunk *fc = chunks[cur_data.fc_idx];
-  if (!fc->is_mine()) abort("wrong fields chunk");
-  fc->sources[ft] = tmp->add_to(fc->sources[ft]);
-  require_component(c);
+  if (!fc->is_mine()) meep::abort("wrong fields chunk");
+
+  fc->add_source(ft, src_vol(c, src, std::move(index_arr), std::move(amplitudes)));
+  // We can't do require_component(c) since that only works if all processes are adding
+  // srcdata for the same components in the same order, which may not be true.
+  // ... instead, the caller should call fields::require_source_components()
+  //     after all add_srcdata calls are complete.
 }
 
-static realnum *amp_func_data_re = NULL;
-static realnum *amp_func_data_im = NULL;
+static double *amp_func_data_re = NULL;
+static double *amp_func_data_im = NULL;
 static const volume *amp_func_vol = NULL;
 static size_t amp_file_dims[3];
 
@@ -395,20 +379,20 @@ void fields::add_volume_source(component c, const src_time &src, const volume &w
   std::string dataset_im = std::string(dataset) + ".im";
 
   size_t re_dims[] = {1, 1, 1};
-  double *real_data = eps_file.read(dataset_re.c_str(), &rank, re_dims, 3);
+  realnum *real_data = (realnum *)eps_file.read(dataset_re.c_str(), &rank, re_dims, 3, sizeof(realnum) == sizeof(float));
   if (verbosity > 0)
     master_printf("read in %zdx%zdx%zd amplitude function file \"%s:%s\"\n", re_dims[0], re_dims[1],
                   re_dims[2], filename, dataset_re.c_str());
 
   size_t im_dims[] = {1, 1, 1};
-  double *imag_data = eps_file.read(dataset_im.c_str(), &rank, im_dims, 3);
+  realnum *imag_data = (realnum *)eps_file.read(dataset_im.c_str(), &rank, im_dims, 3, sizeof(realnum) == sizeof(float));
   if (verbosity > 0)
     master_printf("read in %zdx%zdx%zd amplitude function file \"%s:%s\"\n", im_dims[0], im_dims[1],
                   im_dims[2], filename, dataset_im.c_str());
 
   for (int i = 0; i < 3; ++i) {
     if (re_dims[i] != im_dims[i]) {
-      abort("Imaginary and real datasets have different dimensions");
+      meep::abort("Imaginary and real datasets have different dimensions");
     }
   }
 
@@ -429,11 +413,11 @@ void fields::add_volume_source(component c, const src_time &src, const volume &w
                                complex<double> A(const vec &), complex<double> amp) {
   volume where(where_); // make a copy to adjust size if necessary
   if (gv.dim != where.dim)
-    abort("incorrect source grid_volume dimensionality in add_volume_source");
+    meep::abort("incorrect source grid_volume dimensionality in add_volume_source");
   LOOP_OVER_DIRECTIONS(gv.dim, d) {
     double w = user_volume.boundary_location(High, d) - user_volume.boundary_location(Low, d);
     if (where.in_direction(d) > w + gv.inva)
-      abort("Source width > cell width in %s direction!\n", direction_name(d));
+      meep::abort("Source width > cell width in %s direction!\n", direction_name(d));
     else if (where.in_direction(d) > w) { // difference is less than 1 pixel
       double dw = where.in_direction(d) - w;
       where.set_direction_min(d, where.in_direction_min(d) - dw * 0.5);
@@ -485,7 +469,7 @@ static std::complex<double> gaussianbeam_ampfunc(const vec &p) {
   case Hx: return EH[3];
   case Hy: return EH[4];
   case Hz: return EH[5];
-  default: abort("invalid component in gaussianbeam_ampfunc");
+  default: meep::abort("invalid component in gaussianbeam_ampfunc");
   }
 }
 
@@ -499,7 +483,7 @@ void fields::add_volume_source(const src_time &src, const volume &where, gaussia
             ? 0
             : where.in_direction(Y) == 0 ? 1 : where.in_direction(Z) == 0 ? 2 : -1;
     if (n == -1)
-      abort(
+      meep::abort(
           "can't determine source direction for non-empty source volume with NO_DIRECTION source");
   }
   bool has_tm = abs(beam.get_E0(2)) > 0;
@@ -524,7 +508,7 @@ void fields::add_volume_source(const src_time &src, const volume &where, gaussia
 
 gaussianbeam::gaussianbeam(const vec &x0_, const vec &kdir_, double w0_, double freq_,
                            double eps_, double mu_, std::complex<double> E0_[3]) {
-  if (x0.dim == Dcyl) abort("wrong dimensionality in gaussianbeam");
+  if (x0_.dim == Dcyl) meep::abort("wrong dimensionality in gaussianbeam");
 
   x0 = x0_;
   kdir = kdir_;
